import csv
import sqlite3
from functools import wraps
from typing import List, Any, Union

from arca.ext.database.imports import *


def none_as_null(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """
        Replace None as null()
        """
        func(self, *args, **kwargs)
        for k, v in self.__dict__.items():
            if v is None:
                setattr(self, k, null())

    return wrapper


def map_attributes(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """
        Map kwargs to class attributes
        """
        func(self, *args, **kwargs)
        for k, v in kwargs.items():
            if getattr(self, k):
                setattr(self, k, v)

    return wrapper


def false_if_error(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """
        Return False if error encountered
        """
        try:
            return func(self, *args, **kwargs)
        except Exception as error_occurred:
            return False

    return wrapper


def table_exists(table_name: str) -> Any:
    def decorator(func):
        def wrapper(*args, **kwargs):
            if table_name not in args[0].registered_tables:
                raise Exception(f"Table schema {table_name} does not exist")
            return func(*args, **kwargs)

        return wrapper

    return decorator


class SQLAlchemyDatabase:
    def __init__(self,
                 sqlite_file: str,
                 table_schemas: List):
        self.sqlite_file = sqlite_file

        self.engine = None
        self.base = None
        self.meta = None
        self.session = None

        self.url = f'sqlite:///{sqlite_file}'

        self.setup()

        self.registered_tables = []
        for schema in table_schemas:
            schema.__table__.create(self.engine, checkfirst=True)
            self.registered_tables.append(schema.__tablename__)

    def export_table_to_csv(self, table_name: str, file_path: str) -> bool:
        try:
            conn = sqlite3.connect(self.sqlite_file)
            cursor = conn.cursor()
            cursor.execute(f"SELECT * FROM {table_name}")

            with open(file_path, 'w', newline='') as csvfile:
                writer = csv.writer(csvfile)
                writer.writerow([i[0] for i in cursor.description])
                writer.writerows(cursor.fetchall())
            conn.close()
            return True
        except Exception as error_occurred:
            return False

    def commit(self):
        self.session.commit()

    def close(self):
        self.session.close()

    def setup(self):
        if not self.url:
            return

        self.engine = create_engine(self.url)

        if not self.engine:
            return

        if not database_exists(self.engine.url):
            create_database(self.engine.url)

        self.base = declarative_base(bind=self.engine)
        self.meta = MetaData()
        self.meta.create_all(self.engine)

        Session = sessionmaker()
        Session.configure(bind=self.engine)
        self.session = Session()

    def get_first_entry(self, table_schema, order=None) -> Union[None, Any]:
        query = self.session.query(table_schema)
        if order is not None:
            query = query.order_by(order)
        return query.first()

    def get_all_entries(self, table_schema, order=None) -> List[Any]:
        query = self.session.query(table_schema)
        if order is not None:
            query = query.order_by(order)
        return query.all()

    def get_all_by_filters(self, table, order=None, **kwargs) -> List[Any]:
        query = self.session.query(table).filter_by(**kwargs)
        if order is not None:
            query = query.order_by(order)
        return query.all()

    def get_first_by_filters(self, table, order=None, **kwargs) -> Any:
        query = self.session.query(table).filter_by(**kwargs)
        if order is not None:
            query = query.order_by(order)
        return query.first()

    def delete_by_filters(self, table, order=None, **kwargs) -> bool:
        query = self.session.query(table).filter_by(**kwargs)
        if order is not None:
            query = query.order_by(order)
        query.delete()
        self.commit()
        return True

    def get_attribute_from_first_entry(self, table_schema, field_name, order=None) -> Any:
        entry = self.get_first_entry(table_schema=table_schema, order=order)
        return getattr(entry, field_name, None)

    def set_attribute_of_first_entry(self, table_schema, field_name, field_value, order=None) -> bool:
        entry = self.get_first_entry(table_schema=table_schema, order=order)
        if not entry:
            entry = self.create_entry(table_schema, **{field_name: field_value})
        return self.update_entry_single_field(entry, field_name, field_value)

    @false_if_error
    def create_entry(self, table_schema, **kwargs) -> Union[bool, Any]:
        entry = table_schema(**kwargs)
        self.session.add(entry)
        self.commit()
        return entry

    @false_if_error
    def create_entry_if_does_not_exist(self, table_schema, fields_to_check: List[str], **kwargs) -> Union[bool, Any]:
        filters = {k: v for k, v in kwargs.items() if k in fields_to_check}
        entries = self.get_all_by_filters(table=table_schema, **filters)
        if not entries:
            return self.create_entry(table_schema=table_schema, **kwargs)
        return entries[0]

    @false_if_error
    def update_entry_single_field(self, entry, field_name, field_value) -> bool:
        setattr(entry, field_name, field_value)
        self.commit()
        return True

    @false_if_error
    def update_entry_multiple_fields(self, entry, **kwargs) -> bool:
        for field, value in kwargs.items():
            setattr(entry, field, value)
        self.commit()
        return True


class CustomTable:
    def __init__(self):
        self._ignore = []



from sqlalchemy import create_engine, null, MetaData, Table, Column, Integer, String, Boolean, BigInteger
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, DeclarativeMeta
from sqlalchemy_utils import database_exists, create_database
from enum import Enum



class ArcaFilePath:
    def __init__(self, cog_name: str, file_path: str):
        self.path = f"cogs/{cog_name}/{file_path}"

    def __str__(self):
        return self.path

    def __repr__(self):
        return self.path

    def __eq__(self, other):
        return self.path == other.path

    def __hash__(self):
        return hash(self.path)



from typing import List

import yaml
import mergedeep
from mergedeep import Strategy


class ConfigParser:
    def __init__(self, name: str, config_files: List[str]):
        self.name = name
        self.config = {}
        for config_file in config_files:
            with open(config_file, "r") as f:
                data = yaml.load(f, Loader=yaml.SafeLoader)
                mergedeep.merge(self.config, data, strategy=Strategy.REPLACE)

    def get(self, key: str, path: List[str] = None, default=None):
        try:
            if path is None:
                path = []
            value = self.config
            for p in path:
                value = value[p]
            return value.get(key, default)
        except KeyError:
            return default



from typing import List

import discord
from discord.ext.commands import Cog

from arca.base.bot_config import ArcaConfig
from arca.base.cog_config import ArcaCogConfig
from arca.discord.slash_commands import SlashCommand, SlashCommandGroup
from arca.discord.utilities import get_roles, get_users


class ArcaCog(Cog):
    def __init__(self, bot: discord.Bot, config: ArcaCogConfig, bot_config: ArcaConfig = None):
        self.bot = bot
        self.config = config
        self.bot_config = bot_config

    def register_slash_commands(self, commands: List[SlashCommand]):
        for command in commands:
            self.bot.application_command(name=command.name, cls=discord.SlashCommand, parent=command.group)(
                command.func)

    def register_slash_command_groups(self, groups: List[SlashCommandGroup]):
        for group in groups:
            self.bot.add_application_command(group)

    @property
    def admin_role_names(self) -> List[str]:
        admin_role_names = self.config.admin_role_names
        if not admin_role_names and self.bot_config:
            admin_role_names = self.bot_config.admin_role_names
        return admin_role_names

    def get_admin_roles(self, guild: discord.Guild) -> List[discord.Role]:
        return get_roles(guild=guild, role_names=self.admin_role_names)

    @property
    def admin_ids(self) -> List[int]:
        admin_ids = self.config.admin_ids
        if not admin_ids and self.bot_config:
            admin_ids = self.bot_config.admin_ids
        return admin_ids

    def get_admin_users(self, guild: discord.Guild) -> List[discord.Member]:
        return get_users(guild=guild, user_ids=self.admin_ids)



from typing import List

from discord import Bot, Guild, Role, Member

from arca.base.bot_config import ArcaConfig
from arca.discord.utilities import get_roles, get_users


class Arca(Bot):
    def __init__(self, config: ArcaConfig, *args, **options):
        super().__init__(*args, **options)
        self.config = config

    @property
    def server_id(self) -> int:
        return self.config.server_id

    @property
    def token(self) -> str:
        return self.config.token

    @property
    def prefix(self) -> str:
        return self.config.prefix

    @property
    def admin_role_names(self) -> List[str]:
        return self.config.admin_role_names

    def get_admin_roles(self, guild: Guild) -> List[Role]:
        return get_roles(guild=guild, role_names=self.admin_role_names)

    @property
    def admin_ids(self) -> List[int]:
        return self.config.admin_ids

    def get_admin_users(self, guild: Guild) -> List[Member]:
        return get_users(guild=guild, user_ids=self.admin_ids)



import sys
from contextlib import contextmanager
from typing import Union, List, Optional

import discord
from discord.ext.bridge import BridgeContext
from discord.ext.commands import Context


# Custom error messages

@contextmanager
def disable_exception_traceback():
    """
    All traceback information is suppressed and only the exception type and value are printed
    """
    default_value = getattr(sys, "tracebacklimit", 1000)  # `1000` is a Python's default value
    sys.tracebacklimit = 0
    yield
    sys.tracebacklimit = default_value  # revert changes


class PermissionAbuseWarning(Exception):
    """
    A warning message.
    """

    def __init__(self, user_name: str, message: Optional[str] = None):
        super().__init__(message or f"{user_name} attempted to run a command that they do not have permission to run.")


# Custom functions

async def send_error(ctx: Union[BridgeContext, Context], error_message: Optional[str] = None):
    await ctx.respond(error_message or "Something went wrong!")


def get_roles(guild: discord.Guild, role_names: List[str]) -> List[discord.Role]:
    filtered_roles = []
    for role in guild.roles:
        if role.name in role_names:
            filtered_roles.append(role)
    return filtered_roles


def get_users(guild: discord.Guild, user_ids: List[int]) -> List[discord.Member]:
    members = []
    for member in guild.members:
        if member.id in user_ids:
            members.append(member)
    return members


def get_users_with_roles(guild: discord.Guild, role_names: List[str]) -> List[discord.Member]:
    members = []
    roles = get_roles(guild, role_names)
    for member in guild.members:
        if any(x in member.roles for x in roles):
            members.append(member)
    return members


def get_users_without_roles(guild: discord.Guild, role_names: List[str]) -> List[discord.Member]:
    members = []
    roles = get_roles(guild, role_names)
    for member in guild.members:
        if not any(x in member.roles for x in roles):
            members.append(member)
    return members


def user_has_role(user: discord.Member, role_name: str) -> bool:
    role = discord.utils.get(user.guild.roles, name=role_name)
    return role in user.roles


def user_does_not_have_role(user: discord.Member, role_name: str) -> bool:
    role = discord.utils.get(user.guild.roles, name=role_name)
    return role not in user.roles


async def user_is_admin(cog: 'BaseCog', ctx: Union[BridgeContext, Context]) -> bool:
    # TODO: improve speed on this later
    # check IDs first, most likely less of them, but more strict
    admin_users = cog.get_admin_users(guild=ctx.guild)
    if admin_users and ctx.author in admin_users:
        return True
    # check admin roles next
    admin_roles = cog.get_admin_roles(guild=ctx.guild)
    if admin_roles and any(x in ctx.author.roles for x in admin_roles):
        return True
    return False




import logging
from typing import List

from arca.internal.config_parser import ConfigParser


class ArcaConfig(ConfigParser):
    def __init__(self, bot_name: str, config_files: List[str] = None):
        if config_files is None:
            config_files = []
        super().__init__(name=bot_name, config_files=config_files)

    @property
    def log_level(self):
        level = self.get(key='Logging', default='INFO')
        return logging.getLevelName(level)

    @property
    def token(self):
        return self.get(key='BotToken', path=['Discord'], default=None)

    @property
    def prefix(self):
        return self.get(key='BotPrefix', path=['Discord'], default=None)

    @property
    def server_id(self):
        return self.get(key='ServerID', path=['Discord'], default=None)

    @property
    def admin_ids(self):
        return self.get(key='AdminIDs', path=["Discord"], default=[])

    @property
    def admin_role_names(self):
        return self.get(key='AdminRoleNames', path=['Discord'], default=[])



from typing import List

from arca.ext.database.imports import *

from arca.utilities import convert_to_bool

Base = declarative_base()


def get_table_schema_name(table: DeclarativeMeta) -> str:
    return getattr(table, "__name__", None)


def get_table_columns(table: Table) -> List[Column]:
    return table.columns._all_columns


def get_table_column_names(table: Table) -> List[str]:
    columns = get_table_columns(table=table)
    return [column.name for column in columns]


def table_schema_to_name_type_pairs(table: Table):
    columns = get_table_columns(table=table)
    pairs = {}
    ignore_columns = getattr(table, "_ignore", [])
    for column in columns:
        if column not in ignore_columns:
            pairs[column.name] = sql_type_to_human_type_string(column.type)
    return pairs


def sql_type_to_human_type_string(sql_type) -> str:
    if not hasattr(sql_type, "python_type"):
        return ""

    python_type = sql_type.python_type
    if python_type == str:
        return "String"
    elif python_type in [int, float]:
        return "Number"
    elif python_type == bool:
        return "True/False"
    return ""


def human_type_to_python_type(human_type: str):
    try:
        return float(human_type)  # is it a float?
    except:
        try:
            return int(human_type)  # is it an int?
        except:
            bool_value = convert_to_bool(bool_string=human_type)
            if bool_value is not None:  # is it a bool?
                return bool_value
            else:
                return human_type  # it's a string



import logging
from typing import List

from arca import ArcaFilePath
from arca.internal.config_parser import ConfigParser


class ArcaCogConfig(ConfigParser):
    def __init__(self, cog_name: str, config_files: List[str] = None):
        if not config_files:
            config_files = []

        adjusted_config_files = []
        for config_file in config_files:
            adjusted_config_files.append(ArcaFilePath(cog_name=cog_name, file_path=config_file).path)
        super().__init__(name=cog_name, config_files=config_files)

    @property
    def log_level(self):
        level = self.get(key='Logging', default='INFO')
        return logging.getLevelName(level)

    @property
    def admin_ids(self):
        return self.get(key='AdminIDs', path=["Discord"], default=[])

    @property
    def admin_role_names(self):
        return self.get(key='AdminRoleNames', path=["Discord"], default=[])



from functools import wraps

from dbotbase.discord_utils import user_is_admin, send_error, disable_exception_traceback, PermissionAbuseWarning


# Custom decorators

def is_admin(func):
    @wraps(func)
    async def inner(*args, **kwargs):
        cog = args[0]
        ctx = args[1]
        if await user_is_admin(cog=cog, ctx=ctx):
            return await func(*args, **kwargs)
        await send_error(ctx=ctx, error_message="You are not an admin!")
        with disable_exception_traceback():
            raise PermissionAbuseWarning(user_name=ctx.author.name)

    return inner



import setuptools

__version__ = '0.2.0'

__title__ = "arcacog"
__author__ = 'Nate Harris'
__author_email__ = 'n8gr8gbln@gmail.com'
__github_username__ = "nwithan8"
__github_repo__ = "arcacog"
__copyright__ = "Copyright © 2022 - Nate Harris"
__license__ = 'GNU General Public License v3 (GPLv3)'
__description__ = "A package to build cogs for Arca"
__keywords__ = ["Discord", "bot", "cogs", "base", "Arca"]

with open("README.md", "r") as fh:
    long_description = fh.read()

REQUIREMENTS = [
    "base",
]

setuptools.setup(
    name=__title__,
    packages=setuptools.find_packages(),
    version=__version__,
    license=__license__,
    description=__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=__author__,
    author_email=__author_email__,
    url=f'https://github.com/{__github_username__}/{__github_repo__}',
    download_url=f'https://github.com/{__github_username__}/{__github_repo__}/archive/refs/tags/{__version__}.tar.gz',
    keywords=__keywords__,
    install_requires=REQUIREMENTS,
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Development Status :: 4 - Beta',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3',  # Specify which python versions that you want to support
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Multimedia',
        'Topic :: Internet :: WWW/HTTP',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3.8, <4'
)



from typing import List


def is_positive_int(n):
    return n.isdigit()


def convert_to_bool(bool_string: str):
    """
    Careful: True or False is valid. Check if is None to see if this conversion failed
    """
    if bool_string.lower() in ['false', 'no', 'off', 'disable']:
        return False
    elif bool_string.lower() in ['true', 'yes', 'on', 'enable']:
        return True
    return None


def int_to_place(n: int) -> str:
    """
    Convert a number to a place.
    """
    last_digit = int(repr(n)[-1])
    last_two_digits = int(repr(n)[-2:])
    suffix = 'th'
    if last_digit == 1 and last_two_digits != 11:
        suffix = 'st'
    elif last_digit == 2 and last_two_digits != 12:
        suffix = 'nd'
    elif last_digit == 3 and last_two_digits != 13:
        suffix = 'rd'
    return f"{n}{suffix}"


class DynamicClassCreator:
    def __init__(self):
        self.created_classes = {}

    def __call__(self, bases: List):
        rep = ",".join([i.__name__ for i in bases])
        if rep in self.created_classes:
            return self.created_classes[rep]

        class DynamicClass(*bases):
            pass

        self.created_classes[rep] = DynamicClass
        return DynamicClass
